@@ -80,11 +80,23 @@ class ConsoleApplication(object):
     def __init__(self, run_until_return=await_enter, on_stop=None):
         plain_terminal = os.environ.get("TERM", "").lower() == "none"
 
+        # Windows doesn't have SIGPIPE
+        if hasattr(signal, 'SIGPIPE'):
+            signal.signal(signal.SIGPIPE, signal.SIG_DFL)
+
         colorama.init(strip=True if plain_terminal else None)
 
         parser = OptionParser(usage=self._usage(), version=frida.__version__)
 
         if self._needs_device():
+            relays = []
+            def add_relay(option, opt_str, value, parser):
+                try:
+                    address, username, password, kind = value.split(",")
+                    relays.append(frida.Relay(address, username, password, kind))
+                except:
+                    parser.error("Invalid relay, schema is: “address,username,password,turn-{udp,tcp,tls}”")
+
             parser.add_option("-D", "--device", help="connect to device with the given ID",
                     metavar="ID", type='string', action='store', dest="device_id", default=None)
             parser.add_option("-U", "--usb", help="connect to USB device",
@@ -93,11 +105,29 @@ class ConsoleApplication(object):
                     action='store_const', const='remote', dest="device_type", default=None)
             parser.add_option("-H", "--host", help="connect to remote frida-server on HOST",
                     metavar="HOST", type='string', action='store', dest="host", default=None)
+            parser.add_option("--certificate", help="speak TLS with HOST, expecting CERTIFICATE",
+                    metavar="CERTIFICATE", type='string', action='store', dest="certificate", default=None)
+            parser.add_option("--origin", help="connect to remote server with “Origin” header set to ORIGIN",
+                    metavar="ORIGIN", type='string', action='store', dest="origin", default=None)
+            parser.add_option("--token", help="authenticate with HOST using TOKEN",
+                    metavar="TOKEN", type='string', action='store', dest="token", default=None)
+            parser.add_option("--keepalive-interval",
+                    help="set keepalive interval in seconds, or 0 to disable (defaults to -1 to auto-select based on transport)",
+                    metavar="INTERVAL", type='int', action='store', dest="keepalive_interval", default=None)
+            parser.add_option("--p2p", help="establish a peer-to-peer connection with target",
+                    action='store_const', const='p2p', dest="session_transport", default="multiplexed")
+            parser.add_option("--stun-server", help="set STUN server ADDRESS to use with --p2p",
+                    metavar="ADDRESS", type='string', action='store', dest="stun_server", default=None)
+            parser.add_option("--relay", help="add relay to use with --p2p",
+                    metavar="address,username,password,turn-{udp,tcp,tls}", type='string',
+                    action='callback', callback=add_relay)
 
         if self._needs_target():
             def store_target(option, opt_str, target_value, parser, target_type, *args, **kwargs):
                 if target_type == 'file':
                     target_value = [target_value]
+                elif target_type == 'gated':
+                    target_value = re.compile(target_value)
                 setattr(parser.values, 'target', (target_type, target_value))
             parser.add_option("-f", "--file", help="spawn FILE", metavar="FILE",
                 type='string', action='callback', callback=store_target, callback_args=('file',))
@@ -107,6 +137,8 @@ class ConsoleApplication(object):
                 type='string', action='callback', callback=store_target, callback_args=('name',))
             parser.add_option("-p", "--attach-pid", help="attach to PID", metavar="PID",
                 type='int', action='callback', callback=store_target, callback_args=('pid',))
+            parser.add_option("-W", "--await", help="await spawn matching PATTERN", metavar="PATTERN",
+                type='string', action='callback', callback=store_target, callback_args=('gated',))
             parser.add_option("--stdio", help="stdio behavior when spawning (defaults to “inherit”)", metavar="inherit|pipe",
                 type='choice', choices=['inherit', 'pipe'], default='inherit')
             parser.add_option("--aux", help="set aux option when spawning, such as “uid=(int)42” (supported types are: string, bool, int)", metavar="option",
@@ -136,15 +168,30 @@ class ConsoleApplication(object):
             self._device_id = options.device_id
             self._device_type = options.device_type
             self._host = options.host
+            self._certificate = options.certificate
+            self._origin = options.origin
+            self._token = options.token
+            self._keepalive_interval = options.keepalive_interval
+            self._session_transport = options.session_transport
+            self._stun_server = options.stun_server
+            self._relays = relays
         else:
             self._device_id = None
             self._device_type = None
             self._host = None
+            self._certificate = None
+            self._origin = None
+            self._token = None
+            self._keepalive_interval = None
+            self._session_transport = 'multiplexed'
+            self._stun_server = None
+            self._relays = None
         self._device = None
         self._schedule_on_output = lambda pid, fd, data: self._reactor.schedule(lambda: self._on_output(pid, fd, data))
         self._schedule_on_device_lost = lambda: self._reactor.schedule(self._on_device_lost)
         self._spawned_pid = None
         self._spawned_argv = None
+        self._selected_spawn = None
         self._target_pid = None
         self._session = None
         if self._needs_target():
@@ -177,14 +224,14 @@ class ConsoleApplication(object):
             target = getattr(options, 'target', None)
             if target is None:
                 if len(args) < 1:
-                    parser.error("target file, process name or pid must be specified")
+                    parser.error("target must be specified")
                 target = infer_target(args[0])
                 args.pop(0)
             target = expand_target(target)
             if target[0] == 'file':
                 argv = target[1]
                 argv.extend(args)
-            args = []
+                args = []
             self._target = target
         else:
             self._target = None
@@ -253,6 +300,9 @@ class ConsoleApplication(object):
             return
         if self._spawned_pid is not None:
             self._device.resume(self._spawned_pid)
+            if self._target[0] == 'gated':
+                self._device.disable_spawn_gating()
+                self._device.off('spawn-added', self._on_spawn_added)
         self._resumed = True
 
     def _exit(self, exit_status):
@@ -269,20 +319,48 @@ class ConsoleApplication(object):
                 self._update_status("Device '%s' not found" % self._device_id)
                 self._exit(1)
                 return
+        elif (self._host is not None) or (self._device_type == 'remote'):
+            host = self._host
+
+            options = {}
+            if self._certificate is not None:
+                options['certificate'] = self._certificate
+            if self._origin is not None:
+                options['origin'] = self._origin
+            if self._token is not None:
+                options['token'] = self._token
+            if self._keepalive_interval is not None:
+                options['keepalive_interval'] = self._keepalive_interval
+
+            if host is None and len(options) == 0:
+                self._device = frida.get_remote_device()
+            else:
+                self._device = frida.get_device_manager().add_remote_device(host if host is not None else "127.0.0.1",
+                                                                            **options)
         elif self._device_type is not None:
             self._device = find_device(self._device_type)
             if self._device is None:
                 return
-        elif self._host is not None:
-            self._device = frida.get_device_manager().add_remote_device(self._host)
         else:
             self._device = frida.get_local_device()
         self._device.on('output', self._schedule_on_output)
         self._device.on('lost', self._schedule_on_device_lost)
         if self._target is not None:
+            target_type, target_value = self._target
+
+            if target_type == 'gated':
+                self._device.on('spawn-added', self._on_spawn_added)
+                try:
+                    self._device.enable_spawn_gating()
+                except Exception as e:
+                    self._update_status("Failed to enable spawn gating: %s" % e)
+                    self._exit(1)
+                    return
+                self._update_status("Waiting for spawn to appear...")
+                return
+
             spawning = True
             try:
-                target_type, target_value = self._target
                 if target_type == 'frontmost':
                     try:
                         app = self._device.get_frontmost_application()
@@ -315,12 +393,7 @@ class ConsoleApplication(object):
                     if not self._quiet:
                         self._update_status("Attaching...")
                 spawning = False
-                self._target_pid = attach_target
-                self._session = self._device.attach(attach_target, realm=self._realm)
-                if self._enable_debugger:
-                    self._session.enable_debugger()
-                    self._print("Chrome Inspector server listening on port 9229\n")
-                self._session.on('detached', self._schedule_on_session_detached)
+                self._attach(attach_target)
             except frida.OperationCancelledError:
                 self._exit(0)
                 return
@@ -334,6 +407,24 @@ class ConsoleApplication(object):
         self._start()
         self._started = True
 
+    def _attach(self, pid):
+        self._target_pid = pid
+
+        self._session = self._device.attach(pid, realm=self._realm)
+        self._session.on('detached', self._schedule_on_session_detached)
+
+        if self._session_transport == 'p2p':
+            peer_options = {}
+            if self._stun_server is not None:
+                peer_options['stun_server'] = self._stun_server
+            if self._relays is not None:
+                peer_options['relays'] = self._relays
+            self._session.setup_peer_connection(**peer_options)
+
+        if self._enable_debugger:
+            self._session.enable_debugger()
+            self._print("Chrome Inspector server listening on port 9229\n")
+
     def _show_message_if_no_device(self):
         if self._device is None:
             self._print("Waiting for USB device to appear...")
@@ -342,6 +433,41 @@ class ConsoleApplication(object):
         self._reactor.cancel_io()
         self._exit(0)
 
+    def _on_spawn_added(self, spawn):
+        thread = threading.Thread(target=self._handle_spawn, args=(spawn,))
+        thread.start()
+
+    def _handle_spawn(self, spawn):
+        pid = spawn.pid
+
+        pattern = self._target[1]
+        if pattern.match(spawn.identifier) is None or self._selected_spawn is not None:
+            self._print(Fore.YELLOW + Style.BRIGHT + "Ignoring: " + str(spawn) + Style.RESET_ALL)
+            try:
+                self._device.resume(pid)
+            except:
+                pass
+            return
+
+        self._selected_spawn = spawn
+
+        self._print(Fore.GREEN + Style.BRIGHT + "Handling: " + str(spawn) + Style.RESET_ALL)
+        try:
+            self._attach(pid)
+            self._reactor.schedule(lambda: self._on_spawn_handled(spawn))
+        except Exception as e:
+            error = e
+            self._reactor.schedule(lambda: self._on_spawn_unhandled(spawn, error))
+
+    def _on_spawn_handled(self, spawn):
+        self._spawned_pid = spawn.pid
+        self._start()
+        self._started = True
+
+    def _on_spawn_unhandled(self, spawn, error):
+        self._update_status("Failed to handle spawn: %s" % error)
+        self._exit(1)
+
     def _on_output(self, pid, fd, data):
         if pid != self._target_pid or data is None:
             return
