@@ -56,7 +56,9 @@ def main():
                                           style=style,
                                           history=history,
                                           completer=self._completer,
-                                          complete_in_thread=True)
+                                          complete_in_thread=True,
+                                          enable_open_in_editor=True,
+                                          tempfile_suffix=".js")
                 self._dumb_stdin_reader = None
             else:
                 self._cli = None
@@ -72,6 +74,8 @@ def main():
                               type='string', action='store', dest="user_parameters", default=None)
             parser.add_option("-C", "--cmodule", help="load CMODULE", metavar="CMODULE",
                               type='string', action='store', dest="user_cmodule", default=None)
+            parser.add_option("--toolchain", help="CModule toolchain to use when compiling from source code",
+                              metavar="any|internal|external", type='choice', choices=['any', 'internal', 'external'], default='any')
             parser.add_option("-c", "--codeshare", help="load CODESHARE_URI", metavar="CODESHARE_URI",
                               type='string', action='store', dest="codeshare_uri", default=None)
             parser.add_option("-e", "--eval", help="evaluate CODE", metavar="CODE",
@@ -107,10 +111,11 @@ def main():
 
             if options.user_cmodule is not None:
                 self._user_cmodule = os.path.abspath(options.user_cmodule)
-                with codecs.open(self._user_cmodule, 'rb', 'utf-8') as f:
+                with open(self._user_cmodule, 'rb') as f:
                     pass
             else:
                 self._user_cmodule = None
+            self._toolchain = options.toolchain
 
             self._codeshare_uri = options.codeshare_uri
             self._codeshare_script = None
@@ -215,9 +220,13 @@ def main():
             script.on('message', on_message)
             script.load()
 
-            cmodule_source = self._create_cmodule_source()
-            if cmodule_source is not None:
-                script.exports.frida_repl_load_cmodule(cmodule_source)
+            cmodule_code = self._load_cmodule_code()
+            if cmodule_code is not None:
+                # TODO: Remove this hack once RPC implementation supports passing binary data in both directions.
+                if is_byte_array(cmodule_code):
+                    script.post({'type': 'frida:cmodule-payload'}, data=cmodule_code)
+                    cmodule_code = None
+                script.exports.frida_load_cmodule(cmodule_code, self._toolchain)
 
             stage = 'early' if self._target[0] == 'file' and is_first_load else 'late'
             try:
@@ -380,7 +389,7 @@ def main():
    . . . .       object?   -> Display information about 'object'
    . . . .       exit/quit -> Exit
    . . . .
-   . . . .   More info at https://www.frida.re/docs/home/""".format(version=frida.__version__))
+   . . . .   More info at https://frida.re/docs/home/""".format(version=frida.__version__))
 
         def _print_help(self, expression):
             # TODO: Figure out docstrings and implement here. This is real jankaty right now.
@@ -494,7 +503,7 @@ def main():
             return prompt_string
 
         def _evaluate(self, text):
-            result = self._script.exports.frida_repl_evaluate(text)
+            result = self._script.exports.frida_evaluate(text)
             if is_byte_array(result):
                 return ('binary', result)
             elif isinstance(result, dict):
@@ -538,68 +547,96 @@ def main():
 
             if self._user_script is not None:
                 with codecs.open(self._user_script, 'rb', 'utf-8') as f:
-                    user_script += f.read().rstrip("\r\n") + "\n\n// Frida REPL script:\n"
+                    user_script += f.read()
 
-            return "_init();" + user_script + """\
+            if len(user_script) == 0:
+                return "(function () {\n" + self._make_repl_runtime(indent=4) + "\n})();"
+
+            if user_script.startswith("📦\n"):
+                runtime = self._make_repl_runtime(indent=0)
+                raw_runtime = runtime.encode("utf-8")
+                return "📦\n{} /frida/repl.js\n✄\n{}\n✄\n{}".format(len(raw_runtime), runtime, user_script[2:])
+            else:
+                return "_init();" + user_script + """\n\n// Frida REPL script:\n\
 
 function _init() {
-    global.cm = null;
-    global.cs = {};
-
-    const rpcExports = {
-        fridaReplEvaluate(expression) {
-            try {
-                const result = (1, eval)(expression);
-                if (result instanceof ArrayBuffer) {
-                    return result;
-                } else {
-                    const type = (result === null) ? 'null' : typeof result;
-                    return [type, result];
-                }
-            } catch (e) {
-                return ['error', {
-                    name: e.name,
-                    message: e.message,
-                    stack: e.stack
-                }];
-            }
-        },
-        fridaReplLoadCmodule(source) {
-            const cs = global.cs;
-
-            if (cs._frida_log === undefined)
-                cs._frida_log = new NativeCallback(onLog, 'void', ['pointer']);
-
-            global.cm = new CModule(source, cs);
-        },
-    };
-
-    Object.defineProperty(rpc, 'exports', {
-        get() {
-            return rpcExports;
-        },
-        set(value) {
-            for (const [k, v] of Object.entries(value)) {
-                rpcExports[k] = v;
-            }
-        }
-    });
-
-    function onLog(messagePtr) {
-        const message = messagePtr.readUtf8String();
-        console.log(message);
-    }
+    """ + self._make_repl_runtime(indent=4) + """
 }
 """
 
-        def _create_cmodule_source(self):
+        def _make_repl_runtime(self, indent):
+            code = """\
+global.cm = null;
+global.cs = {};
+
+const rpcExports = {
+    fridaEvaluate(expression) {
+        try {
+            const result = (1, eval)(expression);
+            if (result instanceof ArrayBuffer) {
+                return result;
+            } else {
+                const type = (result === null) ? 'null' : typeof result;
+                return [type, result];
+            }
+        } catch (e) {
+            return ['error', {
+                name: e.name,
+                message: e.message,
+                stack: e.stack
+            }];
+        }
+    },
+    fridaLoadCmodule(code, toolchain) {
+        const cs = global.cs;
+
+        if (cs._frida_log === undefined)
+            cs._frida_log = new NativeCallback(onLog, 'void', ['pointer']);
+
+        if (code === null) {
+            recv('frida:cmodule-payload', (message, data) => {
+                code = data;
+            });
+        }
+
+        global.cm = new CModule(code, cs, { toolchain });
+    },
+};
+
+Object.defineProperty(rpc, 'exports', {
+    get() {
+        return rpcExports;
+    },
+    set(value) {
+        for (const [k, v] of Object.entries(value)) {
+            rpcExports[k] = v;
+        }
+    }
+});
+
+function onLog(messagePtr) {
+    const message = messagePtr.readUtf8String();
+    console.log(message);
+}
+"""
+
+            if indent != 0:
+                indent_str = indent * " "
+                return "\n".join([indent_str + line for line in code.split("\n")])
+            else:
+                return code
+
+        def _load_cmodule_code(self):
             if self._user_cmodule is None:
                 return None
 
-            name = os.path.basename(self._user_cmodule)
+            with open(self._user_cmodule, 'rb') as f:
+                code = f.read()
+            if code_is_native(code):
+                return code
+            source = code.decode('utf-8')
 
-            with codecs.open(self._user_cmodule, 'rb', 'utf-8') as f:
-                source = f.read()
+            name = os.path.basename(self._user_cmodule)
 
             return """static void frida_log (const char * format, ...);\n#line 1 "{name}"\n""".format(name=name) + source + """\
 #line 1 "frida-repl-builtins.c"
@@ -759,6 +796,9 @@ URL: {url}
                             before_dot = '0.0' + before_dot
                         elif t[0] == Token.Punctuation and t[1] == ']':
                             before_dot = '[]' + before_dot
+                        elif t[0] == Token.Punctuation and t[1] == ')':
+                            # we don't know the returned value of the function call so we abort the completion
+                            return
 
                     break
 
@@ -851,6 +891,16 @@ URL: {url}
         def iterbytes(data):
             return (ord(char) for char in data)
 
+    OS_BINARY_SIGNATURES = set([
+        b"\x4d\x5a",         # PE
+        b"\xca\xfe\xba\xbe", # Fat Mach-O
+        b"\xcf\xfa\xed\xfe", # Mach-O
+        b"\x7fELF",          # ELF
+    ])
+
+    def code_is_native(code):
+        return (code[:4] in OS_BINARY_SIGNATURES) or (code[:2] in OS_BINARY_SIGNATURES)
+
     app = REPLApplication()
     app.run()
 
